""" a series of converter to transform *gluon.dal* objects into Ext JS widgets.

"""
import gluon.dal
import re

from extjs import *
from gluon import current
from gridmodifier import ROW_NUMBERING
from helper import (encode_field,
                    get_field_validators,
                    get_foreign_field,
                    get_set_field,
                    get_store_id,
                    is_foreign_field,
                    is_set_field)
from mapper import map_default

# Associated gluon.dal.field type with an ExtJS widget
# The dictionary contains configuration parameters of the Ext JS widget
FTYPE_TO_XTYPE = {'boolean':      (CheckBox, {}), \
                  'date':         (FieldDate, {'format': 'Y-m-d'}), \
                  'datetime':     (FieldDate, {'format': 'Y-m-d H:i:s'}), \
                  'double':       (FieldNumber, {'decimalPrecision':2,
                                                 'decimalSeparator': '.'}), \
                  'id':           (FieldText, {}), \
                  'integer':      (FieldNumber, {}), \
                  'json':         (FieldDict, {}),
                  'list:string':  (FieldList, {}), \
                  'list:integer': (FieldTextArea, {}), \
                  'password':     (FieldText, {'inputType': 'password'}), \
                  'reference':    (ComboBox, {}), \
                  'string':       (FieldText, {}), \
                  'text':         (FieldTextArea, {}), \
                  'time':         (FieldTime, {}), \
                  'upload':       (FieldText, {'inputType': 'file'})}


# constant defining a ExtJS store
ROOT = 'records'
NOPAGE = ''
PAGE = 'page'
SUCCESS = 'success'
TOTAL = 'count'

# xtype defining linked comboBoxes
X_LINKED_COMBO = ['xcomboboxmaster', 'xcomboboxslave']


def _to_field(field, linkedcombo=True, **kwargs):
    """Build the :class:`.Field` configuration associated
    to the database ``field``. The conversion takes into account the
    :class:`.FieldsModifier` instructions
    but does not handle :class:`.FieldContainer`.

    Args:
        field (gluon.dal.Field): database field

        linkedcombo (bool): ``LinkedComboBox`` are ignored when ``False``.
            Useful to build grid filter.

    Keyword Args:
        any Ext JS configuration parameters of the
        ``Ext.form.Field`` class. They are applied in the following order:
        constructor, modifiers and keyword arguments.

    Returns:
        Field:
            The return configurator is a :class:`.Field`.
            However it can be a  :class:`.ComboBox`,
            a :class:`.FieldDate`,
            a :class:`.FieldTextArea`, ...
            depending on the ``type`` of the database field.
            The association is defined in the dictionary
            :const:`FTYPE_TO_XTYPE`.

    .. attention::
        It is recommended to used the more general method :meth:`to_field`
        which also understands :class:`.FieldContainer`.

    """
    T = current.T

    fieldname = field.name
    tablename = field.tablename

    f_name = encode_field(tablename, fieldname)

    # foreign key
    if is_foreign_field(field):

        k_tablename, k_fieldname, k_key = get_foreign_field(field)

        cfg = ComboBox(name=f_name,
                       hiddenName=f_name,
                       displayField=encode_field(k_tablename, k_fieldname),
                       valueField=encode_field(k_tablename, k_key),
                       store=get_store_id(k_tablename))

    # set field. Use the basic Ext JS ComboBox
    elif is_set_field(field):
        cfg = dict(name=f_name,
                   store=get_set_field(field),
                   xtype='combobox')

    # other types of fields
    else:
        cls, di = FTYPE_TO_XTYPE[field.type]
        cfg = cls(name=f_name, **di)

    # field label translate in the local language
    cfg["fieldLabel"] = str(T(field.label))

    # default value
    if field.default:
        cfg["value"] = field.default

        # the checkBox is rendered initially checked
        if field.type == 'boolean':
            cfg["checked"] = True

    # not null
    if field.notnull:
        cfg["allowBlank"] = False
        cfg["afterLabelTextTpl"] = '<span style="color:red"> * </span>'

    # read only field
    if field.readable and not field.writable:
        cfg["readOnly"] = True

    # tool tip
    if field.comment:
        cfg["tipText"] = field.comment

    # validator constraints
    cfg.update(get_field_validators(field))

    # hide primary key
    # hide field which are not readable and not writable
    # hide field request by the form modifier
    hidden = field.name == "id"
    hidden = hidden or ((not field.readable) and (not field.writable))

    modifier_forms = PluginManager('dbui').dbui.modifier_forms
    if tablename in modifier_forms:
        if fieldname in modifier_forms[tablename].hidden_fields:
            hidden = True

    if hidden:
        cfg["hidden"] = True
        cfg["hideLabel"] = True
        cfg["readOnly"] = True

    # configuration options set by the field_modifers
    # at this stage parameter for LinkedComboBox are applied
    modifier_fields = PluginManager('dbui').dbui.modifier_fields
    if tablename in modifier_fields:
        if fieldname in modifier_fields[tablename].extjs_fields:

            extjs = modifier_fields[tablename].extjs_fields[fieldname]

            process = True
            if not linkedcombo and 'xtype' in extjs:
                process = not (extjs['xtype'] in X_LINKED_COMBO)

            if process:
                cfg.update(extjs)

    # configuration options from the keyword arguments
    cfg.update(kwargs)

    return cfg


def to_field(field, composite=True, linkedcombo=True, **kwargs):
    """Build the :class:`.Field` configuration for
    the database field. The conversion handles
    :class:`.FieldContainer` and takes into account
    the :class:`.FieldsModifier` instructions.

    .. note::
        The return configurator is a :class:`.Field`.
        However it can be a :class:`.ComboBox`,
        a :class:`.FieldDate`,
        :class:`.FieldTextArea`, ...
        depending on the ``type`` of the database field.
        The association is defined in the dictionary
        :const:`FTYPE_TO_XTYPE`.

    .. note::
        The database fields can be organized in
        :class:`.FieldContainer` in the application
        model using the modifiers
        :class:`.FieldsModifier`.

    Args:
        field (gluon.dal.Field);

        composite (bool): ``FieldContainer`` is ignored when ``False``.
            Useful to build grid filter.

        linkedcombo (bool): ``LinkedComboBox`` are ignored when ``False``.
            Useful to build grid filter.

    Keyword Args:
        kwargs (dict):
            any Ext JS configuration parameter of the ``Ext.form.Field``
            class. They are applied in the following order:
            constructor, modifiers and keyword arguments.

    Returns:
        * Field
        * ``None`` when the argument composite is ``True`` and when
            the field is consumed by a ``FieldContainer``.

    """
    table = field.table
    tablename = field.tablename
    modifier_fields = PluginManager('dbui').dbui.modifier_fields

    # do we have composite field for this table ?
    composite_fields = None
    if tablename in modifier_fields:
        modifier_field = modifier_fields[tablename]
        composite_fields = modifier_field.composite_fields

    # main field of the composite field
    # it will consume the embedded field too
    if composite and composite_fields and field.name in composite_fields.main:

        cfg = FieldContainer()

        for fieldname in composite_fields[field.name].fields:
            cfg.append_items(_to_field(table[fieldname]))

        # use configuration options from the modifier
        cfg.update(composite_fields[field.name].extjs)

    # an embedded field already used in a composite field
    elif composite and composite_fields and field.name in composite_fields.others:
        cfg = None

    # a basic field
    else:
        cfg = _to_field(field, linkedcombo=linkedcombo)

    # configuration options from keyword arguments
    if cfg:
        cfg.update(kwargs)

    return cfg


def to_fields(table):
    """Build the list :class:`.Field` configuration
    for each field of the database ``table``.

    The function handles :class:`.FieldSet` and
    :class:`.FieldContainer` as well as the
    :class:`.FieldsModifier` instructions.
    The ``FieldSet`` and ``FieldContainer`` can be defined in the
    application model using the modifier
    :class:`.FormModifier` and the
    :class:`.FieldsModifier` respectively.

    Args:
        table (gluon.dal.Table): database table

    Returns:
        list:
            The list elements are :class:`.Field`, but
            it can be a :class:`.ComboBox`,
            a :class:`.FieldDate`,
            a :class:`.FieldTextArea`, ...,
            depending on the ``type`` of the database field.

    """
    li = []
    tablename = table._tablename
    modifier_forms = PluginManager('dbui').dbui.modifier_forms

    # do we have FieldSets
    field_sets, modifier_form = None, None
    if tablename in modifier_forms:
        modifier_form = modifier_forms[tablename]
        if modifier_form and modifier_form.field_sets:
            field_sets = modifier_form.field_sets

    # Table with Ext.form.FieldSet and/or Ext.form.Fieldcontainer
    # and/or Ext.form.Field and/or Spacer

    # NOTE: the field id is add in order to run this form
    # in all context, create, update, destroy.
    # It is add to the latest fieldset.
    id_is_used = False

    if field_sets:
        for fieldset in field_sets:

            cfg = FieldSet()
            for fieldname in fieldset.fields:

                if isinstance(fieldname, Base):
                    cfg.append_items(fieldname)
                    continue

                if fieldname == 'id':
                    id_is_used = True

                field = table[fieldname]
                item = to_field(field)
                if item:
                    cfg.append_items(item)

            # fieldset configuration options ser by the modifier
            cfg.update(fieldset.extjs)
            li.append(cfg)

        # append the field Id the last field set
        if not id_is_used:
            li[-1].append_items(to_field(table['id']))

    # Table with Ext.form.Fieldcontainers and/or Ext.form.Fields
    else:
        for field in table:
            cfg = to_field(field)
            if cfg:
                li.append(cfg)

    # map the list of fields/fieldSets on Ext.form.formPanel
    mapper = map_default
    if modifier_form and modifier_form.mapper:
        mapper = modifier_form.mapper

    li = mapper(li)

    return li


def to_formPanel(table, **kwargs):
    """Build the :class:`.FormPanel` configuration
    for the database ``table``. The conversion takes into account the
    :class:`.FormModifier` instructions.

    Args:
        table (gluon.dal.Table): database table

    Keyword Args:
        kwargs (dict):
            any Ext JS configuration parameter of the
            ``Ext.form.Panel`` class. They are
            applied in the following order: constructor, modifiers
            and keyword arguments.

    Returns:
        FormPanel:

    """
    tablename = table._tablename

    fields = to_fields(table)
    cfg = FormPanel(items=fields)

    # add a store for non dummy database
    if table._db._uri:
        cfg['store'] = get_store_id(tablename)

    # configuration options from form modifier
    modifier_forms = PluginManager('dbui').dbui.modifier_forms
    if tablename in modifier_forms:
        cfg.update(modifier_forms[tablename].extjs)

    # configuration options form the keyword arguments
    cfg.update(kwargs)

    return cfg


def to_gridColumn(field, **kwargs):
    """Build the :class:`.GridColumn` configuration
    for the database ``field``. The conversion takes into account the
    :class:`.GridModifier` instructions.

    Args:
        field (gluon.dal.Field): database field

    Keyword Args:
        kwargs (dict):
            any Ext JS configuration parameter of the
            ``Ext.grid.Column`` class.
            They are applied in the following order: constructor, modifiers
            and keyword arguments.

    Returns:
        GridColumn:

    """
    T = current.T

    fieldname = field.name
    tablename = field.tablename

    cfg = GridColumn(header=str(T(field.label)),
                     sortable=True)

    # name of the field in the data store
    # replace foreign key by the pointing column
    if is_foreign_field(field):
        k_tablename, k_fieldname, k_key = get_foreign_field(field)
        index = encode_field(k_tablename, k_fieldname)

    else:
        index = encode_field(tablename, fieldname)

    cfg["dataIndex"] = index

    # hide the primary key
    if fieldname == "id":
        cfg["hidden"] = True

    # hide field which are not readable and not writable
    if (not field.readable) and (not field.writable):
        cfg["hidden"] = True

    # dedicated column render fo json field
    if field.type == "json":
        cfg["xtype"] = "jsoncolumn"

    # Hide fields request by the grid modifiers
    modifier_grids = PluginManager('dbui').dbui.modifier_grids
    if tablename in modifier_grids:
        if fieldname in modifier_grids[tablename].hidden_columns:
            cfg["hidden"] = True

    # configuration options from the grid modifier
    if tablename in modifier_grids:
        if fieldname in modifier_grids[tablename].configure_columns:
            cfg.update(modifier_grids[tablename].configure_columns[fieldname])

    # configuration options from the keyword arguments
    cfg.update(kwargs)

    return cfg


def to_gridColumnModel(table):
    """Build the :class:`.GridColumnModel` configuration
    for the database ``table``. The conversion takes into account the
    :class:`.GridModifier` instructions.

    Args:
        table (gluon.dal.Table): database table

    Returns:
        GridColumnModel:

    """
    delete_columns = []
    template_columns = []
    tablename = table._tablename
    row_numbering = ROW_NUMBERING

    # get modifier requirements
    modifier_grids = PluginManager('dbui').dbui.modifier_grids
    if tablename in modifier_grids:
        delete_columns = modifier_grids[tablename].delete_columns
        template_columns = modifier_grids[tablename].template_columns
        row_numbering = modifier_grids[tablename].row_numbering

    # standard column
    cfg = GridColumnModel()
    for field in table:
        fieldname = field.name
        if fieldname not in delete_columns:
            cfg.append(to_gridColumn(field))

    # template column
    for tpl in template_columns:
        col = GridTemplateColumn(sortable=True)
        col.update(tpl.extjs)
        cfg.insert(tpl.position, col)

    # row numbering in the first column
    if row_numbering:
        cfg.insert(0, GridRowNumberer())

    return cfg


def to_gridFilter(table, **kwargs):
    """Build the :class:`.GridFilter` configuration for
    the database ``table``. The ``GridFilter`` is parametrized using the
    :class:`.GridModifier`.

    Args:
        table (gluon.dal.Table): database table

    Keyword Args:
        kwargs (dict):
            any of the Ext JS configuration parameters of the
            ``App.grid.Filter`` class. They are applied in the following order:
            constructor, modifiers, keyword arguments.

    A filter rule is defined in the model via the
    :class:`.GridModifier`.
    The rule is a ``tuple`` containing 3 strings
    ``("field1", "operator", "comment")``
    where ``field1`` is the name of a field belonging to the ``table``.
    A more elaborate filter rule allows to filter on any foreign field using
    the syntax is ``("table2.field1", "operator", "comments")``.
    Valid operators are defined in the method
    :meth:`.DbSvc._encode_query`.

    Rules are store in the modifier section ``grid_filters``.

    Returns:
        GridFilter:
            Return an empty dictionary if the GridFilter is
            not defined in the application model.


    """
    T = current.T
    tablename = table._tablename

    modifier_grids = PluginManager('dbui').dbui.modifier_grids
    if tablename not in modifier_grids:
        return {}

    grid_filters = modifier_grids[tablename].grid_filters
    if not grid_filters:
        return {}

    if not grid_filters.filters:
        return {}

    di = GridFilter(title=T('Filter %s' % tablename))

    for i in range(len(grid_filters.filters)):

        fieldname, operator, comment = grid_filters.filters[i]

        # filter rule dealing with any foreign field
        m = re.match(r"(\w+)\.(\w+)", fieldname)
        if m:
            ktablename = m.group(1)
            kfieldname = m.group(2)
            field = table._db[ktablename][kfieldname]
            name = '[%s.%s]' % (ktablename, kfieldname)

        # basic filter rule
        else:
            field = table[fieldname]
            name = '[%s.%s]' % (tablename, fieldname)

        # get the standard configuration for the field
        # ignore composite field set up for form
        # ignore linked ComboBox set up for form
        cfg = to_field(field, composite=False, linkedcombo=False)

        # remove field constraints making no sense for a filter
        if 'allowBlank' in cfg:
            cfg["allowBlank"] = True

        if 'readOnly' in cfg:
            cfg['readOnly'] = False

        if 'value' in cfg:
            del cfg['value']

        # Remove the star in the label for required field
        if field.notnull and 'afterLabelTextTpl' in cfg:
            del cfg['afterLabelTextTpl']

        # replace all textarea by textfield
        if cfg['xtype'] == 'xtextarea':
            cfg['xtype'] = 'xtextfield'

        # combobox can be reset individually
        # not valid for combobox associated to a set of data
        if cfg['xtype'] == 'xcombobox' and not is_set_field(field):
            cfg['emptyText'] = T('select...')
            cfg['xtype'] = 'xcomboboxuserreset'

        # prepare the name for the where close [table.field]
        cfg['name'] = name

        # store information to customise the filter
        cfg['filterComment'] = comment
        cfg['filterOperator'] = operator
        cfg['filterType'] = field.type

        # copy the comment to be used as a tool tip
        # when the plugin pFormToolTip is used
        cfg['tipText'] = comment

        # field configuration option
        cfg.update(grid_filters.extjs_filters[i])

        di.append_items(cfg)

    # configuration options from the grid modifier
    di.update(grid_filters.extjs)

    # configuration option from keyword arguments
    di.update(kwargs)

    return di


def to_gridPanel(table, **kwargs):
    """Build the :class:`.GridPanel` configuration for
    the database ``table``. The conversion takes into account the
    :class:`.GridModifier` instructions.

    Args:
        table (gluon.dal.Table): database table

    Keyword Args:
        kwargs (dict):
            any of the Ext JS configuration parameters of the
            ``App.grid.Panel`` class. They are applied in the following order:
            constructor, modifiers, keyword arguments.

    Returns:
        GridPanel:

    """
    tablename = table._tablename

    cfg = GridPanel(columns=to_gridColumnModel(table),
                    frame=False,
                    viewConfig={'forceFit': True})

    # add a store for non dummy database
    if table._db._uri:
        cfg['store'] = get_store_id(tablename)

    # configuration option from the grid modifier
    modifier_grids = PluginManager('dbui').dbui.modifier_grids
    if tablename in modifier_grids:
        cfg.update(modifier_grids[tablename].extjs)

    # configuration option from the keyword arguments
    cfg.update(kwargs)

    # grid with filter
    filter = to_gridFilter(table)
    if filter:
        cfg = GridWithFilter(panelItems=dict(cfg),
                             selectorItems=filter)

        if tablename in modifier_grids:
            cfg.update(modifier_grids[tablename].grid_with_filter.extjs)

    return cfg


def to_jsonstore(table, **kwargs):
    """Build the :class:`.DirectStore` configuration for
    the database ``table``. The conversion takes into account the
    :class:`.StoreModifier` instructions.

    Args:
        table (gluon.dal.Table): database table

    Keyword Args:
        kwargs (dict):
            any of the Ext JS configuration parameters of the
            ``App.data.DirectStore`` class. They are applied in the following order:
            constructor, modifiers, keyword arguments.

    Returns:
        DirectStore:

    """
    db = table._db
    tablename = table._tablename

    # NOTE: the configuration option extraParams defined the
    # transactions parameters received by the server.
    #
    # This dictionary contains 3 keys:
    #    tablename: name of the table in the database
    #    dbFields: list of fields which should be red
    #    where: define inner join, mainly to resolve foreign key
    #
    # NOTE: the keyword pageParam allow to activate / inhibit pagination
    # It should be equal ro "page" or undefined respectively.
    base_params = {'tableName': tablename, 'dbFields': []}

    cfg = DirectStore(autoLoad=False,
                      autoSync=True,
                      extraParams=base_params,
                      model=tablename,
                      pageParam=NOPAGE,
                      root=ROOT,
                      successProperty=SUCCESS,
                      storeId=get_store_id(tablename),
                      totalProperty=TOTAL)

    for field in table:
        fieldname = field.name

        # the field of the table
        base_params['dbFields'].append([tablename, fieldname])

        # NOTE: the store contains all fields of the table.
        # It also contains the pointing values for foreign keys
        # Therefore the store has the the id and the value for
        # each foreign key. This technique is use to resolve foreign key
        # in form and in grid.
        if is_foreign_field(field):

            if 'where' not in base_params: base_params['where'] = []

            # the foreign field
            k_tablename, k_fieldname, k_key = get_foreign_field(field)
            base_params['dbFields'].append([k_tablename, k_fieldname])

            # define the where close
            nt = (tablename, fieldname, k_tablename, k_key)
            base_params['where'].append("[%s.%s] == [%s.%s]" % nt)

    # configuration options from the store modifier
    # namely the order by directive
    modifier_stores = PluginManager('dbui').dbui.modifier_stores
    if tablename in modifier_stores:

        base_params['orderby'] = []
        for el in modifier_stores[tablename].orderby:

            if isinstance(el, gluon.dal.Field):
                field, dir = el, 'ASC'

            elif isinstance(el, gluon.dal.Expression):
                field, dir = el.first, 'DESC'

            base_params['orderby'].append([field.tablename, field.name, dir])

        # Ext JS configuration option from the modifier
        cfg.update(modifier_stores[tablename].extjs)


    # Ext JS configuration options from the keyword arguments
    cfg.update(kwargs)

    return cfg


def to_model(table):
    """Build the :class:`.Model` configuration for
    the database ``table``.

    Args:
        table (gluon.dal.Table): database table

    Returns:
        dict:

    """
    db = table._db
    tablename = table._tablename

    cfg = Model(fields=[])

    for field in table:
        fieldname = field.name

        # the field of the table
        # encode field name and define default value
        field_cfg = {'name': encode_field(tablename, fieldname)}
        if field.default:
            field_cfg['defaultValue'] = field.default

        cfg['fields'].append(field_cfg)

        # NOTE: the store contains all fields of the table.
        # It also contains the pointing values for foreign keys
        # Therefore the store has the the id and the value for
        # each foreign key. This technique is use to resolve foreign key
        # in form and in grid.
        if is_foreign_field(field):

            k_tablename, k_fieldname, k_key = get_foreign_field(field)

            # encode the field name
            k_field_cfg = {'name': encode_field(k_tablename, k_fieldname)}
            cfg['fields'].append(k_field_cfg)

            # retrieve the default value for the foreign key
            if field.default:
                row = db(db[k_tablename].id == field.default).select().first()
                if row:
                    k_field_cfg['defaultValue'] = row[k_fieldname]

    return cfg


def to_panelWithUrlSelector(table, selectorTitle='Select', **kwargs):
    """Build the :class:`.PanelWithUrlSelector` configuration
    where the selector is a form derived from the database ``table``.

    Args:
        table (gluon.dal.Table): database table

        selectorTitle (str):
            the title of the FieldSet encapsulating
            the selector fields.

        baseUrl (str): the selected values are send to a controller
            defined by the baseUrl.Values are processed and results
            will be published in the panel.
            It is mandatory to specified the baseUrl.

    Keyword Args:
        kwargs (dict):
            any of the Ext JS configuration parameters of the
            ``App.panel.WithUrlSelector`` class.

    Returns:
        PanelWithUrlSelector:

    :raise BaseException: when baseUrl is not defined.

    """
    if 'baseUrl' not in kwargs:
        raise BaseException('The keyword argument basseUrl is missing.')

    # build the selector from the table
    fields = to_fields(table)
    selector = FieldSet(items=fields,
                        plugins=['pFormToolTip'],
                        title=current.T(selectorTitle))

    # build the configuration for the PanelWithUrlSelector
    default = dict(selectorItems=selector)

    default.update(kwargs)
    cfg = PanelWithUrlSelector(**default)

    return cfg


def to_treeNodes():
    """Build the list of :class:`.Node`.
    The Node structure is defined in the application model using
    :class:`.ViewportModifier`.

    Returns:
        list:

    """
    cfg = []

    modifier_viewports = PluginManager('dbui').dbui.modifier_viewports

    if modifier_viewports:
        nav_nodes = modifier_viewports.nodes
        for node in nav_nodes:
            cfg.append(node.get_node())

    return cfg


def to_viewport(**kwargs):
    """Build the :class:`.Viewport` configuration.
    The conversion takes into account the
    :class:`.ViewportModifier` instructions.

    Keyword Args:
        kwargs (dict):
            any of the Ext JS configuration parameters of the
            ``App.viewport.Viewport`` class.

    Returns:
        dict:

    """
    cfg = {}

    modifier_viewports = PluginManager('dbui').dbui.modifier_viewports

    if modifier_viewports:
        cfg = modifier_viewports.extjs

        if modifier_viewports.default:
            cfg['defaultTabPath'] = modifier_viewports.default

    # configuration options from the keyword arguments
    cfg.update(kwargs)

    return cfg
