""" implementation of the ``Ext.Direct`` protocol on the server-side.

"""

import datetime
import json
import pprint
import traceback
import sys

from basesvc import BaseSvc
from gluon import current
from gluon.http import HTTP
from gluon.restricted import RestrictedError
from helper import as_list

DBUI = 'Dbui'
PROC_KEY = '%s.%s'


class MyJsonEncoder(json.JSONEncoder):
    """Add date, datetime and time to the standard JSON encoder.

    """
    def default(self, obj):

        if isinstance(obj, (datetime.date,
                            datetime.datetime,
                            datetime.time)):
            return obj.isoformat().replace('T', ' ')

        return json.JSONEncoder.default(self, obj)



class DirectSvc(BaseSvc):
    """Generic service to implement the
    `Ext.Direct <http://www.sencha.com/products/extjs/extdirect/>`_ protocol
    on the server-side.

    This implementation is based on the
    `web2py service <http://www.web2py.com/book/default/chapter/09>`_
    using a simplified version of the class ``gluon.tools.Service``.

    To register functions and make them available on the client-side
    use decorator::

        service = DirectSvc()

        @service.register
        def myfunction(a, b):

    The same technique can be used with methods.

    On the client-side:

        * the API definition can be retrieved using the
          URL: ``/myapplication/plugin_dbui/call/get_api``

        * the URL to send requests is
          ``/myapplication/plugin_dbui/call``

        * remote functions and methods can be call in the
          following way ``Dbui.myfunction(a, b, callback)``
          and ``MyClass.mymethod(c, d, callback)`` respectively.

    In this implementation requests and response are JSON encoded.

    """

    def __init__(self):

        BaseSvc.__init__(self)

        self.procedures = {}
        self.register(self.echo)


    def __call__(self):
        """Check the client request and launch the execution
        of the function / method on the server.

        Returns:
            str:
                * the response of the function / method run
                  on the server side
                * the response is encoded as a JSON string.

        :raise gluon.http.HTTP:
            when the request has no arguments
            or when the function / method crashed.

        """
        self.dbg('Start directSvc.call')
        myrequest = current.request

        # no arguments in a Json-encoded raw HTPP request
        if len(myrequest.args) == 0:
            return self.route()

        self.error(404, 'Object does not exist.')


    def echo(self, x):
        """Return the input arguments to run echo tests.

        """
        return x


    def error(self, code, message):
        """Build and raise the gluon.http.HTTP exception

        Args:
            code (int): code of the HTTP error
            message (str): explanation for the HTTP error

        :raise gluon.http.HTTP:

        """
        raise HTTP(code, message)


    def register(self, f):
        """ Register a function/method.

        Example::

            service = DirectSvc()

            @service.register
            def myfunction(a, b):

        Args:
            f (function): reference to a function

        Returns:
            function:
                the input argument

        """
        # a procedure is registered as 'action.method'
        # The action is equal to Dbui for a function
        # and to the name of the class for a method
        action = DBUI
        if 'im_class' in dir(f):
            action = f.im_class.__name__

        k = PROC_KEY % (action, f.__name__)
        self.procedures[k] = f

        self.dbg("Procedure %s registered" % k)
        return f


    def route(self):
        """Route the request to the appropriate function / method,
        pass the proper arguments and return the results.

        Returns:
            str:
                the response of the function / method
                encoded as a JSON string.

        :raise gluon.http.HTTP:
            with the code 500 when the execution of the
            function / method crash. The python trace back is stored
            in the web2py ticket system.

        """
        self.dbg('Start directSvc.route')

        # decode transactions, a list of Json-encoded raw HTPP requests:
        # {action: "x", method: "y", data: [], type:"rpc", tid:z}
        myrequest = current.request
        transactions = as_list(json.loads(myrequest.body.read()))

        # process each transactions and encode each result as a dict
        # {action: "x", method: "y", result: [], type:"rpc", tid:z}
        li = []
        for arg in transactions:
            di = dict(arg)
            self.dbg('transaction:', di)
            try:
                k = PROC_KEY % (arg['action'], arg['method'])
                if arg['data']:
                    di['result'] = self.procedures[k](*arg['data'])
                else:
                    di['result'] = self.procedures[k]()

            # catch exceptions and log them in the ticker system
            # inform the user that something went wrong in the server
            except:
                r = RestrictedError(layer='directsvc.py',
                                    code='route',
                                    output='',
                                    environment=current.globalenv)
                r.log(myrequest)
                self.error(500, 'Internal server error.')

            del di['data']
            li.append(di)

        self.dbg('End directSvc.call', li)
        return json.dumps(li, cls=MyJsonEncoder)
