""" configurators for the Ext JS components.

"""

from gluon import current
from gluon.storage import Storage
from gluon.tools import PluginManager

MSG_XTYPE = 'Invalid keyword "xtype"'


class ExtJSException(BaseException): pass


class Base(Storage):
    """Base class for Ext JS configurator.

    The base class comes with two methods ``append_items``
    and ``append_plugins``. They are required to append ``items``
    and ``plugins``.

    .. note::
        A protection is set to avoid changing the ``xtype`` of the element.

    """
    xtype = None

    def __init__(self, **kwargs):
        """

        Keyword Args:
            kwargs (dict):
                any Ext JS configuration parameter of the target class

        :raise ExtJSException:
            if the keyword ``xtype`` is in the keyword arguments.

        """
        if 'xtype' in kwargs:
            raise ExtJSException(MSG_XTYPE)

        self['xtype'] = self.xtype

        items = kwargs.pop('items', None)
        if items:
            self._append('items', items)

        plugins = kwargs.pop('plugins', None)
        if plugins:
            self._append('plugins', plugins)

        self.update(kwargs)


    def _append(self, key, t):
        li = list(t)
        if key in self:
            self[key].extend(li)
        else:
            self[key] = li


    def append_items(self, *args):
        self._append('items', args)


    def append_plugins(self, *args):
        self._append('plugins', args)


class ArrayStore(Base):
    """Configurator for the ``Ext.data.ArrayStore``."""
    xtype = 'arraystore'


class CheckBox(Base):
    """Configurator for ``Ext.form.field.Checkbox``."""
    xtype = 'checkbox'


class ComboBox(Base):
    """Configurator for ``App.form.field.CombBox``."""
    xtype = 'xcombobox'


class DirectStore(Base):
    """Configurator for ``App.data.DirectStore``."""
    xtype = 'xdirectstore'


class Field(Base):
    """Configurator for ``Ext.form.field.Field``."""
    xtype = 'field'


class FieldAceEditor(Base):
    """Configurator for the ``App.form.field.AceEditor``."""
    xtype = 'xaceeditorfield'


class FieldContainer(Base):
    """Configurator for ``Ext.form.FieldContainer``."""
    xtype = 'fieldcontainer'


class FieldDate(Base):
    """Configurator for ``Ext.form.field.Date``."""
    xtype = 'datefield'


class FieldDict(Base):
    """Configurator for ``App.form.field.Dict``."""
    xtype = 'xdictfield'


class FieldList(Base):
    """Configurator for ``App.form.field.List``."""
    xtype = 'xlistfield'


class FieldNumber(Base):
    """Configurator for ``Ext.form.field.Number``."""
    xtype = 'numberfield'


class FieldSet(Base):
    """Configurator for ``Ext.form.FieldSet``."""
    xtype = 'fieldset'


class FieldText(Base):
    """Configurator for ``Ext.form.field.Text``.

    :note: Regular expression for the validation can be written as a string.

    """
    xtype = 'textfield'

    def __init__(self, **kwargs):
        self['plugins'] = ['pRegExp']
        Base.__init__(self, **kwargs)


class FieldTextArea(Base):
    """Configurator for ``Ext.form.field.TextArea``.

    .. note::
        Regular expression for the validation can be written as a string.

    """
    xtype = 'textarea'

    def __init__(self, **kwargs):
        self['plugins'] = ['pRegExp']
        Base.__init__(self, **kwargs)


class FieldTime(Base):
    """Configurator for ``Ext.form.field.Time``."""
    xtype = 'timefield'


class FormPanel(Base):
    """Configurator for ``App.form.Panel``."""
    xtype = 'xform'


class GridColumn(dict):
    """Configurator for ``Ext.grid.Column``."""


class GridColumnModel(list):
    """Configurator for ``Ext.grid.ColumnModel``.

    .. note::
        It is  implemented as a ``list`` of :class:`.GridColumn`.

    """


class GridFilter(Base):
    """Configurator for ``App.grid.Filter``."""
    xtype = 'xgridfilter'


class GridPanel(Base):
    """Configurator for ``App.grid.Panel``."""
    xtype = 'xgrid'


class GridRowNumberer(Base):
    """Configurator for ``Ext.grid.column.RowNumberer``."""
    xtype = 'rownumberer'


class GridTemplateColumn(Base):
    """Configurator for ``Ext.grid.TemplateColumn``."""
    xtype = 'templatecolumn'


class GridWithFilter(Base):
    """Configurator for ``App.panel.GridWithFilter``."""
    xtype = 'xgridwithfilter'


class JsonStore(Base):
    """Configurator for the ``Ext.data.JsonStore``."""
    xtype = 'jsonstore'


class Model(dict):
    """Configurator for the ``Ext.data.Model``. """

    def __init__(self, **kwargs):
        dict.__init__(self, **kwargs)
        self['extend'] = 'Ext.data.Model'


class Panel(Base):
    """Configurator for ``Ext.panel.Panel``."""
    xtype = 'panel'


class PanelWithUrlSelector(Base):
    """Configurator for ``App.panel.WithUrlSelector``."""
    xtype = 'xpanelwithurlselector'


class Store(Base):
    """Configurator for the ``Ext.data.Store``."""
    xtype = 'store'


class TabPanel(Base):
    """Configurator for ``Ext.tab.Panel``."""
    xtype = 'tabpanel'


class Viewport(Base):
    """Configurator for ``App.viewport.Viewport``."""
    xtype = 'xviewport'


class Window(Base):
    """Configurator for ``Ext.Window``."""
    xtype = 'window'


class XmlStore(Base):
    """Configurator for the ``Ext.data.XmlStore``."""
    xtype = 'xmlstore'
