""" a series of helper functions

"""

import os
import re
import subprocess
import tempfile

from gluon import current
from gluon.dal import Row, Table
from gluon.http import HTTP
from gluon.validators import (IS_DATE_IN_RANGE,
                              IS_DATETIME_IN_RANGE,
                              IS_DECIMAL_IN_RANGE,
                              IS_EMPTY_OR,
                              IS_FLOAT_IN_RANGE,
                              IS_IN_DB,
                              IS_IN_SET,
                              IS_INT_IN_RANGE,
                              IS_MATCH,
                              IS_LENGTH)

GIT = '/usr/bin/git'
STOREID = '%sStore'


def as_list(val):
    """Encapsulated ``val`` into a ``list`` when ``val`` is not an
    instance of a ``list``.

    Args:
        val (any): input value

    Returns:
        list:

    """
    if not isinstance(val, (list, tuple)):
         val = [val]
    return val


def close_user_session():
    """Close a user session.
    Logout the user and destroy the session file.

    """
    # logout the user
    auth = current.session.auth
    if auth:
        if auth.logout:
            auth.logout()
        del current.session.auth

    # clean the session
    current.session.forget(current.response)


def decode_field(field):
    """Decode a string in CapitalizeWords into a list
    of words where all words are in lower case.
    The reverse function is :meth:`encode_field`.

    :note:
        useful to decode database field name:
        ``TableField`` or ``TableFieldFunction``

    Args:
        field (str): value

    Returns:
        tuple:

    """
    li = []
    for c in field:
        if c.isupper():
            li.append(c.lower())
        elif li:
            li[-1] += c
    return tuple(li)


def dummy_row(table, value=''):
    """Create a dummy ``Row`` for the ``table``. The row contains keys for
    the table and for the reference tables. Each field are set to ``value``.

    The Row van be inserted in a gluon.dal.Rows::

        rows = db(query).select(orderby=db.table.field)
        empty_row = dummy_row(db.table)
        row.records.append(row)

    Args:
        table (gluon.dal.Table): database table
        value (str): value associated to each field either
            an empty string of None

    Returns:
        gluon.dal.Row:

    """
    db = table._db
    tablename = table._tablename

    row = Row()
    row[tablename] = Row()

    for field in table:
        row[tablename][field.name] = value

        if is_foreign_field(field):
            k_tablename, k_fieldname, k_id = get_foreign_field(field)
            if k_tablename not in row:
                for k, v in dummy_row(db[k_tablename], value=value).iteritems():
                    if k not in row:
                        row[k] = v

    return row


def encode_field(*args):
    """Encode the string arguments into a CapitalizeWords.
    The reverse function is :meth:`decode_field`.

    Returns:
        str:

    .. note::
        useful to build database field name like ``TableField``
        or ``TableFieldFunc``.

    .. note::
        The Ext JS ComboBox don't decode properly any separator between
        table and field like table+field. Therefore choice for the encoding
        are very limited.

    """
    # NOTE: don't use the method title since foo_faa
    # becomes Foo_Faa and you want Foo_faa.
    return ''.join([el[0].upper() + el[1:].lower() for el in args if len(el)])


def encode_validator_errors(error):
    """Encode and translate the validator errors.

    Args:
        error (dict): the key is the table field encode as
            ``TablenameFieldname`` while the value is the
            error message return by the validator

    Returns:
        dict:
            error message are translated and should be more friendly
            since it contains the field name with its error.

    """
    di = {}
    for k, v in error.iteritems():
        tablename, fieldname = decode_field(k)
        di[k] = "%s %s" % (current.T(fieldname.title()), v)
    return di


def get_all_tables(dal):
    """Get the list of names for all tables in the database, ``dal``,
    including alias table.

    Args:
        dal (gluon.dal.DAL): the database

    Returns:
        list:

    .. note::
        It is recommend to use standard iterator provided by
        the gluon.dal.DAL class. However by construction, they ignore
        alias table.

    """
    li = [dal[attr] for attr in dir(dal) if isinstance(dal[attr], Table)]
    return li


def get_create_id(table, **kwargs):
    """Get the ``id`` of the record identified by a set of ``field=value``
    keyword arguments.
    If the record does not exist it is created with the values
    specified in the keywords arguments.

    Example::

        id = get_create_id(db.person, first_name="Joe", last_name="Doe")

    Args:
        table (gluon.dal.Table): database table

    Returns:
        int:
            the ``id`` of the record

    """
    id = get_id(table, **kwargs)

    if id == None:
        id = table.insert(**kwargs)

    return id


def get_field_validators(field):
    """Get the Ext JS configuration parameters to run
    database gluon.dal.Field validators on the client-side.

    Args:
        field (gluon.dal.Field): database field

    Returns:
        dict:
            the keys depend on the validators.
            Possible values are ``minValue``, ``maxValue``,
            ``regex``, ``regexText``,``maxLength``, ``minLength``, ...

    """
    cfg = {}

    validators = as_list(field.requires)

    for el in validators:

        vdt = el
        if isinstance(vdt, IS_EMPTY_OR):
            vdt = vdt.other

        # date and time validators
        if isinstance(vdt, IS_DATE_IN_RANGE):
            format = '%Y-%m-%d'
            cfg['minValue'] = vdt.minimum.strftime(format)
            cfg['maxValue'] = vdt.maximum.strftime(format)

        elif isinstance(vdt, IS_DATETIME_IN_RANGE):
            format = '%Y-%m-%d %H:%M:%S'
            cfg['minValue'] = vdt.minimum.strftime(format)
            cfg['maxValue'] = vdt.maximum.strftime(format)

        # number validators
        elif isinstance(vdt, (IS_DECIMAL_IN_RANGE, IS_FLOAT_IN_RANGE, IS_INT_IN_RANGE)):
            cfg['minValue'] = vdt.minimum
            cfg['maxValue'] = vdt.maximum

        # string validator
        elif isinstance(vdt, IS_MATCH):
            cfg["regex"] = vdt.regex.pattern
            cfg["regexText"] = vdt.error_message

        # others validator
        elif isinstance(vdt, IS_LENGTH):
            cfg["maxLength"] = vdt.maxsize
            cfg["minLength"] = vdt.minsize

    return cfg


def get_file_paths(path, ext=None, alpha=True):
    """get the local paths for files matching the arguments.
    For example, to get the paths of the javascript source files
    located in the local directory ``static/plugin_dbui/src``::

        li = get_file_paths("static/plugin_dbui/src", ext=".js")

    The local paths are relative to the application directory.

    Args:
        path (None, list or str):
            local path defining the target files:

                * ``None``.
                * a ``string`` pointing to one files.
                * a ``list`` of strings pointing to several files.
                * a ``string`` pointing to a directory.
                * a ``list`` of string pointing to files and directory.

        ext (str):
            select only files with the extension ``ext``, *e.g*
            ``.js``, ``.css``, ...

        alpha (bool): sort the paths in alphabetic order

    Returns:
        list:
            local paths for all selected files.

    """
    li = []
    if path == None:
        return li

    server_path, client_path = get_reference_paths()

    for el in as_list(path):
        p = os.path.join(server_path, el)

        if os.path.isfile(p):
            if ext and (not el.endswith(ext)): continue
            li.append(os.path.join(client_path, el))

        elif os.path.isdir(p):
            for file in os.listdir(p):
                pfile = os.path.join(el, file)
                li.extend(get_file_paths(pfile, ext, alpha))

    if alpha:
        li.sort()

    return li


def get_foreign_field(field):
    """Get a tuple with the reference table name, reference field name and
    reference primary key name, for the ``field``.

    Args:
        field (gluon.dal.Field): database field

    Returns:
        tuple:
            * ``(k_tablename, k_fieldname, k_keyname)``

            * ``k_tablename`` is the name of the reference table

            * ``k_fieldname`` is the name of the reference field, *e.g.*
              ``table.field`` S{->} ``k_table.k_field``

            * ``k_keyname`` is the name of the primary key of the reference table.
              It is requires by the left join,
              ``table.field == k_table.k_key``.
              In almost all case ``k_key`` is equal to ``id``.

            * ``None`` for standard field.

    """
    if not is_foreign_field(field):
        return None

    # foreign field validators IS_IN_BD contains all information
    # on the parent key for the left join and on the parent field
    validators = as_list(field.requires)

    # Several syntax are allowed

    # IS_IN_DB is use.
    for vdt in validators:

        if isinstance(vdt, IS_IN_DB):
            # IS_IN_DB contains only the parent field.
            # The key for the left join is the id key of the parent table
            if len(vdt.ks) == 1:
                return (vdt.ktable, vdt.ks[0], 'id')

            # IS_IN_DB contains the key for the join and the parent field
            else:
                return (vdt.ktable, vdt.ks[0], vdt.ks[1])

    # IS_IN_DB is not use.
    # The parent key for the left join is the id key of the parent table
    # the parent field is the first field of the parent table which differs id
    i = field.type.index(' ')
    ktablename = field.type[i:].strip()

    for kfield in field._db[ktablename]:
        if kfield.name != 'id':
            return (ktablename, kfield.name, 'id')

    raise Exception


def get_id(table, **kwargs):
    """Get the ``id`` of the record identified by a set of ``field=value``
    keyword arguments. Return ``None`` if the record does not exist.

    Args:
        table (gluon.dal.Table): database table

    Returns:
        * int -- the ``id`` of the record
        * ``None`` when the record does not exist

     """
    query = None
    for k, v in kwargs.iteritems():
        el = table[k] == v

        if query:
            query = query & (el)
        else:
            query = el

    row = table(query)
    if row:
        return row.id

    return None


def get_language():
    """Get the primary language of the application from the HTML
    language code, *e.g.* ``en``, ``fr``, ....

    Returns:
        str

    """
    lg = current.T.accepted_language

    # HTML tag for language: primary-dialect
    # Extract the primary language and ignore dialect
    m = re.match('([a-z]+)(\-([a-zA-Z]*))?', lg)
    if m:
        return m.group(1)

    return lg


def get_plugin_path(plugin_name):
    """Get the local path of the plugin identified by its name.

    Args:
        plugin_name (str): name of the plugin, *e.g.* ``plugin_dbui``

    Returns:
        str:
            * the path of the first directory starting with ``plugin_name``.
            * the local path is relative to the application directory.
            * ``None`` if the plugin is not found.

    """
    server_path, client_path = get_reference_paths()

    # look for the full name of the plugin directory
    p_static = os.path.join(server_path, 'static')
    for el in os.listdir(p_static):
        if os.path.isdir(os.path.join(p_static, el)):
            if el.startswith(plugin_name):
                return os.path.join(client_path, 'static', el)

    # Nothing found
    return None


def get_reference_paths():
    """Get a tuple with the absolute and local paths:

        * The absolute path is relative to the web2py directory.
          It is useful on the server-side
        * The local path is relative to the applications directory.
          It is useful on the client-side

    Returns:
        tuple:
            ``("applications/myapp", "/myapp")``

    """
    server_path = os.path.join('applications', current.request.application)
    client_path = os.path.join(os.path.sep, current.request.application)

    return (server_path, client_path)


def get_set_field(field):
    """Get the list of the set data for the ``field`` when defines.

    Args:
        field (gluon.dal.Field): database field

    Returns:
        list:

    """
    validators = as_list(field.requires)

    for vdt in validators:
        if isinstance(vdt, IS_IN_SET):
            return vdt.theset

    return []


def get_script_path(plugin):
    """Get the local path of the main javascript script to be
    ran on the client-side::

        import plugin_dbui as dbui
        plugin = plugins.dbui
        path = dbui.get_script_path(plugin)

    The main script can be selected in several ways:
        * via the default script
        * via the URL calling the application
        * via the configuration parameter of the plugin_dbui

    **Via the default value:**
        * the default script is located in ``static/plugin_dbui/main.js``

    **Via the URL:**
        * the URL should contain the argument ``script=foo`` or ``script=foo.js``
        * the script is search in ``app_script_dir``, one of the ``plugin_dbui``
          configuration parameters.

    **Via plugin_dbui configuration:**
        * the script path is defined by the ``plugin_dbui``
          configuration parameters ``app_script``.

    Args:
        plugin (reference): web2py plugin instance

    Returns:
        str:
            the local path of the script file relative to the
            application directory.

    """
    server_path, client_path = get_reference_paths()

    # script name specifies in the URL
    if "script" in current.request.vars:
        script = current.request.vars.script

        if not script.endswith(".js"):
            script = '%s.js' % script

        pdir = os.path.join(server_path, plugin.app_script_dir)

        if script not in os.listdir(pdir):
            raise HTTP(500, 'Request script "%s" does not exist !!!' % script)

        return os.path.join(client_path, plugin.app_script_dir, script)

    # default script for the application
    if plugin.app_script:
        return os.path.join(client_path, plugin.app_script)

    # default script for the plugin
    return os.path.join(client_path, plugin.dbui_script)


def get_store_id(name):
    """Build the store identifier using ``name``.

    Args:
        name (str):

    Returns:
        str:
            ``nameStore``

    """
    return STOREID % name


def get_versions():
    """Get the versions of the application and plugins.

    Returns:
        list of dict:
            each dictionary contains the name of the software
            and its version identifier. The keys are ``code`` and ``version``

    """
    server_path, client_path = get_reference_paths()


    # plugin ace
    ace = current.T('plugin not install')
    path = os.path.join(server_path, 'static', 'plugin_ace')
    if os.path.exists(path):
        ace = current.T('install')

    # plugin dbui
    fn = os.path.join(server_path, 'static', 'plugin_dbui', 'src', 'App.js')

    fi = open(fn, 'rb')
    s = fi.read()
    fi.close()

    m = re.match("(.+    version: ')([\w._-]*)('.+)", s, re.DOTALL)
    dbui = m.group(2)

    # plugin extjs
    extjs = current.T('plugin not install')
    path = os.path.join(server_path, 'static', 'plugin_extjs')
    if os.path.exists(path):
        fn = os.path.join(server_path, 'static',
                                       'plugin_extjs',
                                       'src',
                                       'version',
                                       'Version.js')

        fi = open(fn, 'rb')
        s = fi.read()
        fi.close()

        m = re.match(r".+ version = '(\d+(\.\d+)*)',", s, re.DOTALL)
        if m:
            extjs = m.group(1)

    # plugin mathjax
    mathjax = current.T('plugin not install')
    path = os.path.join(server_path, 'static', 'plugin_mathjax')
    if os.path.exists(path):
        fn = os.path.join(server_path, 'static', 'plugin_mathjax', 'MathJax.js')

        fi = open(fn, 'rb')
        s = fi.read()
        fi.close()

        m = re.match(r'.+MathJax.version="(\d+(\.\d+)*)";', s, re.DOTALL)
        if m:
            mathjax = m.group(1)

    # web2py
    web2py = ''
    val = current.request.env.web2py_version
    if isinstance(val, str):
        m = re.match('\d+\.\d+\.\d+', val)
        if m:
            web2py = m.group()

    elif isinstance(val, (tuple, list)):
        li = [str(el) for el in val if isinstance(el, int)]
        web2py = '.'.join(li)

    # version of the application
    myapp = current.T("unknown")

    if os.path.exists(GIT):
        ref_path = os.getcwd()
        os.chdir(server_path)

        fi = tempfile.TemporaryFile()
        subprocess.call(["git", "describe", "--tags"], stdout=fi)
        fi.seek(0)
        myapp = fi.read().strip('\n')
        fi.close()

        os.chdir(ref_path)

    # prepare the output
    li = [dict(code="web2py", version=web2py),
          dict(code="dbui", version=dbui),
          dict(code="extjs", version=extjs),
          dict(code="mathjax", version=mathjax),
          dict(code="ace", version=ace),
          dict(code=current.request.application, version=myapp)]

    return li


def get_where_query(table):
    """Build the ``gluon.dal.Query`` resolving foreign keys
    for the ``table``.

    Args:
        table (gluon.dal.Table): database table

    Returns:
        * ``gluon.dal.Query``
        *  ``None`` when the table has no foreign key

    """
    db = table._db
    query = None

    for field in table:
        if not is_foreign_field(field):
            continue

        k_table, k_field, k_key = get_foreign_field(field)
        condition = field == db[k_table][k_key]

        if query:
            query = ((query) & (condition))
        else:
            query = condition

    return query


def is_foreign_field(field):
    """Return ``True`` when the ``field`` is a reference field.

    Args:
        field (gluon.dal.Field): database field

    Returns:
        bool:

    """
    return field.type.startswith("reference")


def is_mathjax():
    """Return ``True`` when the plugin mathjax is installed.

    Returns:
        bool:

    """
    return get_plugin_path('plugin_mathjax') != None


def is_set_field(field):
    """Return ``True`` when the ``field`` value
    should matches those of a ``set``.

    Args:
        field (gluon.dal.Field): database field

    Returns:
        bool:

    """
    validators = as_list(field.requires)

    for vdt in validators:
        if isinstance(vdt, IS_IN_SET):
            return True

    return False


def is_table_with_foreign_fields(table):
    """Return ``True`` when the database ``table`` has at least
    one reference field.

    Args:
        table (gluon.dal.Table): database table

    Returns:
        bool:

    """
    for field in table:
        if is_foreign_field(field):
            return True
    return False


def rows_serializer(rows):
    """Serialize a ``gluon.dal.Rows`` as a list of dictionary.
    It handle the general case where field belongs to different
    table and where function are applied.

    Args:
        rows (gluon.dal.Rows): list of database records

    Returns:
        list:
            * ``[{Table1Field1: val1,...}, {Table1Field1: val2,...}, ...]``
            * A dictionary per row.
            * Field are encoded as TableField.
            * Field with an applied function like COUNT(table.field)
              are encoded as TableFieldCount.

    """
    # regular expressions
    # A field name can only contains [a_zA-Z0-9_] according to web2py
    table_field = re.compile(r'[\w_]+\.[\w_]+')
    func_table_field = re.compile(r'([A-Z]+)\(([\w_]+)\.([\w_]+)\)')

    # Map between column name and TableField
    map = {}
    for col in rows.colnames:
        if not table_field.match(col):
            func, table, field = func_table_field.match(col).groups()
            map[col] = encode_field(table, field, func)
        else:
            table, field = col.split(".")
            map[col] = encode_field(table, field)

    # serialize
    li = []
    for row in rows:
        di = {}
        for col in rows.colnames:
            if not table_field.match(col):
                di[map[col]] = row._extra[col]
            else:
                (table, field) = col.split(".")
                if table in row:
                    di[map[col]] = row[table][field]
                else:
                    di[map[col]] = row[field]
        li.append(di)

    return li
