""" the modifier class

"""
from gluon import current
from gluon.storage import Storage
from gluon.tools import PluginManager


MSG_INVALID_TYPE = "Invalid plugin type"


class Modifier(object):
    """Base class to construct Modifier tools.

    """

    def __init__(self, dpname, tablename=None):
        """Create the ``gluon.storage.Storage`` in the plugin dbui to house
        modifier information.

            * The path to access to the Storage is defined by the ``dpname``
              and ``tablename``.
            * The ``Storage`` is linked to the class attribute ``data``.
            * The ``Storage`` can be accessed outside the modifier
              in the following way::

                  p = PluginManager('dbui')
                  modifier = p.dbui[dpname]

              or when the tablename is defined::

                  p = PluginManager('dbui')
                  modifier = p.dbui[dpname][tablename]

            * The ``Storage`` contains the ``extjs`` key to store
              configuration parameter of the associate Ext JS widget.

        Args:
            dpname (str): name of the data point.
            tablename (str):name of the database table.

        """
        p = PluginManager('dbui')

        if tablename:
            if dpname not in p.dbui:
                p.dbui[dpname] = {}

            if tablename not in p.dbui[dpname]:
                p.dbui[dpname][tablename] = Storage(extjs={})

            self.data = p.dbui[dpname][tablename]

        else:
            if dpname not in p.dbui:
                p.dbui[dpname] = Storage(extjs={})

            self.data = p.dbui[dpname]


    def configure(self, **extjs):
        """Set the configuration parameters for the associated Ext JS widget.

        .. note::
            * Existing value of the configuration parameters are replaced by
              those defined in the keyword arguments.

            * The keyword argument ``plugins`` contains a list of plugins.
              The current list of ``plugins`` are extend with the new ones
              defined by the keyword argument ``plugins``.

              If a plugin with the same ptype alsready exists in the list
              it is replaced by the new definition.

            * A plugin is defined by a string containing is ``pType``.

            * A plugin is defined by a dictionary containing the ``pType``
              and the plugin configuration options::

                  {'ptype': foo, 'option1': val,...}

            * The widget associated by default is replaced by a new one
              when the keyword argument ``xtype`` is used.

        """
        T = current.T

        if 'plugins' in extjs:

            plugins = extjs['plugins']

            if isinstance(plugins, str):
                plugins = [plugins]

            if 'plugins' in self.data.extjs:

                # list of existing ptypes
                ptypes = []
                for el in self.data.extjs['plugins']:
                    if isinstance(el, (str, unicode)):
                        ptypes.append(el)
                    elif isinstance(el, dict):
                        ptypes.append(el['ptype'])
                    else:
                        raise TypeError(T(MSG_INVALID_TYPE))

                # append new plugins replacing existing one
                for plugin in plugins:

                    ptype = ""
                    if isinstance(plugin, (str, unicode)):
                        ptype = plugin
                    elif isinstance(plugin, dict):
                        ptype = plugin['ptype']
                    else:
                        raise TypeError(T(MSG_INVALID_TYPE))

                    if ptype in ptypes:
                        i = ptypes.index(ptype)
                        self.data.extjs['plugins'][i] = plugin
                    else:
                        self.data.extjs['plugins'].append(plugin)

            else:
                self.data.extjs['plugins'] = list(plugins)

            del extjs['plugins']

        self.data.extjs.update(extjs)
