""" the ``Node`` class to build navigation tree.

"""

import locale
from gluon import current


class Node(object):
    """The ``node`` is the key element of the navigation tree.
    It contains children (leaves) which are associated to widgets.
    The widget will be display on the panel of the *Viewport*.

    """
    def __init__(self, text):
        """Construct the node.

        Args:
            text (str): name of the node appearing in the viewport

        """
        self.text = text
        self.children = []
        self.hidden = []


    def add_child(self, text, cfg):
        """Add a child (leaf) to the node.

        Args:
            text (str):
                the name of the leaf

            cfg (dict):
                the Ext JS configuration options defining
                the widget associated to the leaf

        """
        di = {'cfg': cfg, 'leaf': True, 'text': text}
        self.children.append(di)


    def add_children(self, leaves, func=None, hidden=[]):
        """Add children (leaves) to the node.

        Args:
            leaves (list): a list of string

            func (function):
                function translating the leaf name into
                the Ext JS configuration options dictionary.
                The latter defines the associated widget.

            hidden (list): list of children to be hidden.

        .. note::
            * Leaf names are translated in the current language
            * Leaf names are sorted in alphabetic order

        """
        # translator
        T = current.T

        # translate and order leaves in alphabetic order
        # according to local setting
        cvt = {}
        for el in leaves:
            cvt[T(el)] = el

        translate_leaves = cvt.keys()
        translate_leaves.sort(cmp=locale.strcoll)

        # fill the node with its children
        for tr_leaf in translate_leaves:
            leaf = cvt[tr_leaf]
            if leaf in hidden:
                continue

            self.add_child(tr_leaf, func(leaf))


    def get_node(self):
        """Get the dictionary defining the node.

        Returns:
            dict:
                * ``text`` (str): name of the node
                * ``children`` (list): list of dictionary containing
                  Ext JS configuration for the widgets associated to the leaves.

        """
        return {'text': self.text, 'children': self.children}


    def sort_children(self):
        """sort children according to alphabetical order.

        """
        di = {}
        for child in self.children:
            di[child['text']] = child

        names = di.keys()
        names.sort(cmp=locale.strcoll)

        self.children = []
        for name in names:
            self.children.append(di[name])
