""" basic tool to build report.

A report relies on three main components:
    - The ``form`` associated to a database table.
      It allows the user to defined requirements.
      The table is built using a virtual database.
    - The ``selector`` running in a controller to decode the user requirements
      and to select records in the database accordingly.
    - A ``view`` to render the selected records.

A powerful mechanism is in place to render the report in different format
like HTML, CSV, .... It relies on the Selector class, on the IFRAME and
on the view::

    selector = Selector(virtdb.myselector, extfield='format')
    iframe = selector.download()

    if iframe:
        return iframe

"""

from datetime import date, datetime, time
from gluon import current
from gluon.html import IFRAME, URL
from gluon.storage import Storage
from helper import (decode_field,
                    get_foreign_field,
                    is_foreign_field,
                    is_table_with_foreign_fields)


class Selector(Storage):
    """Basic tool to build a report:

        * Decode the data send by the selector widget.
        * Build the query for a given table of the database.
        * Select Rows for a given table of the database.

    This class handle table and its reference field.
    The basic query handles the selector constraints
    and the inner join for reference field. Additional
    constraints can also be added via the method :meth:`append_query`.

    All fields of the selector are attributes of this class::

        value = myselector.myfield
        value = myselector['myfield']

    The type of the return value is the type of the field
    defined in the model.

    """
    def __init__(self, table, exclude_fields=()):
        """

        Args:
            table (gluon.dal.Table):
                the table containing the definition of the selector

            exclude_fields (tuple):
                a list of selector field names
                which have to be ignored in query and select operations.

        """

        self._exclude_fields = exclude_fields
        self._extension = None
        self._extra_queries = []
        self._tablename = table._tablename

        # Decode the current request
        for key in current.request.vars:
            t = decode_field(key)

            if len(t) != 2 :
                continue

            fieldname = t[1]

            # ignore field id
            if fieldname == 'id':
                continue

            val = current.request.vars[key]

            if not val:
                self[fieldname] = val
                continue

            # convert the selector value according to field type
            fieldtype = table[fieldname].type

            if fieldtype == 'boolean':
                val = val == 'true' or val == 'True'

            elif fieldtype == 'date':
                ds = datetime.strptime(val, '%Y-%m-%dT%H:%M:%S')
                val = date(ds.year, ds.month, ds.day)

            elif fieldtype == 'datetime':
                val = datetime.strptime(val, '%Y-%m-%dT%H:%M:%S')

            elif fieldtype == 'double':
                val = float(val)

            elif fieldtype == 'integer':
                val = int(val)

            elif fieldtype == 'time':
                ds = datetime.strptime(val, '%H:%M:%S')
                val = time(ds.hour, ds.minute, ds.second)

            self[fieldname] = val


    def append_query(self, query):
        """Add constraints which will be used in the database query
        and database select operations.

        Args:
            query (gluon.dal.Query): database query

        """
        self._extra_queries.append(query)


    def as_dict(self):
        """Return the selector fields and their values.

        Returns:
            dict:

        """
        di = {}
        for field in self.fields():
            di[field] = self[field]

        return di

    def fields(self):
        """Get the list of field names defined in the selector.

        Returns:
            list:

        """
        li = []
        for k in self:
            if not k.startswith('_'):
                li.append(k)
        return li


    def query(self, table):
        """Build the database query for the database table.
        It includes:

            * inner join for reference fields,
            * selector constraints,
            * extra queries added via the :meth:`append_query` method.
            * the ``extfield`` as well as fields in the ``exclude_fields``
              tuple are not take into account in the query.

        Args:
            table (gluon.dal.Table): database table

        Returns:
            gluon.dal.Query:

        """
        db = table._db
        query = None

        # inner join for foreign keys
        if is_table_with_foreign_fields(table):
            for field in table:
                if is_foreign_field(field):
                    k_table, k_field, k_id = get_foreign_field(field)
                    q = field == db[k_table][k_id]
                    if query:
                        query = (query) & (q)
                    else:
                        query = q

        # constraint from the selector
        for fieldname in self.fields():

            if not self[fieldname] or \
               fieldname in self._exclude_fields:
                continue

            q = table[fieldname] == self[fieldname]

            if query:
                query = (query) & (q)
            else:
                query = q

        # append the extra queries
        for q in self._extra_queries:
            query = (query) & (q)

        return query


    def repr_url(self):
        """Get the URL string corresponding to the current request.

        Returns:
            str:

        """
        vars = []

        tablename = self._tablename
        tablename = tablename[0].upper() + tablename[1:]

        for (k, v) in current.request.vars.iteritems():
            if k.startswith(tablename) and len(v):
                vars.append("%s=%s" % (k, v))

        t = (current.request.env.http_host,
             current.request.env.path_info,
             '&'.join(vars))

        url = "http://%s%s?%s" % t
        return url


    def select(self, table, *args, **kwargs):
        """Select the records of the table according to the selector constraints
        and to the extra queries.

            * The ``extfield`` as well as the fields in the ``exclude_fields``
              tuple are not take into account in the underlying query.

            * It returns Rows which contains reference tables and columns.

        Args:
            table (gluon.dal.Table): database table
            args (list): field with operator (avg, count, max, min and sum)

        Keyword Args:
            kwargs (dict):
                The keyword arguments are those of the
                gluon.dal.Set.select method: orderby, groupby, ....

        Returns:
            gluon.dal.Rows:

        """
        db = table._db
        columns = [table.ALL]

        # columns for foreign tables
        if is_table_with_foreign_fields(table):
            for field in table:
                if is_foreign_field(field):
                    k_table, k_field, k_id = get_foreign_field(field)
                    columns.append(db[k_table].ALL)

        # field with operator
        columns.extend(args)

        # query handling selector and extra constraints
        query = self.query(table)

        # retrieve the selected records
        return db(query).select(*columns, **kwargs)
